'use strict';

Object.defineProperty(exports, "__esModule", {
	value: true
});

var _tty = require('tty');

var _tty2 = _interopRequireDefault(_tty);

var _lodash = require('lodash');

var _lodash2 = _interopRequireDefault(_lodash);

var _updateCheck = require('./update-check');

var _updateCheck2 = _interopRequireDefault(_updateCheck);

var _package = require('../../package.json');

var _package2 = _interopRequireDefault(_package);

var _commandProcessor = require('./command-processor');

var commandProcessor = _interopRequireWildcard(_commandProcessor);

var _cli = require('../cli');

var _cli2 = _interopRequireDefault(_cli);

var _settings = require('../../settings');

var settings = _interopRequireWildcard(_settings);

var _when = require('when');

var _when2 = _interopRequireDefault(_when);

var _log = require('../lib/log');

var _log2 = _interopRequireDefault(_log);

var _process = require('process');

var _process2 = _interopRequireDefault(_process);

var _unindent = require('../lib/unindent');

var _unindent2 = _interopRequireDefault(_unindent);

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } else { var newObj = {}; if (obj != null) { for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) newObj[key] = obj[key]; } } newObj.default = obj; return newObj; } }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

class CLI {
	constructor() {
		//process.on('unhandledRejection', this.globalRejectionHandler.bind(this));
		this.rootCategory = this.setupCommandProcessor();
	}

	globalRejectionHandler(reason, promise) {
		_log2.default.fatal(reason);
		_process2.default.exit(-1);
	}

	setupCommandProcessor() {
		const app = this;

		return commandProcessor.createAppCategory({
			description: (0, _unindent2.default)(`
				Welcome to the Particle Command Line Interface!
				Version ${_package2.default.version}
				https://github.com/particle-iot/particle-cli
				
			`),

			// options for yargs
			inherited: {
				options: {
					/*
     args: {
     	config: true
     },
     */
					v: {
						alias: 'verbose',
						count: true,
						description: 'Increases how much logging to display'
					},
					q: {
						alias: 'quiet',
						count: true,
						description: 'Decreases how much logging to display'
					}
					/*
      interactive: {
      boolean: true,
      description: 'forces interactive mode'
      },
      'non-interactive': {
      boolean: true,
      description: 'Run in non-interactive mode. This means all required data must be passed as command line arguments.'
      },
      'json': {
      boolean: true,
      description: 'Output in JSON format instead of human friendly'
      },
      */
				}
			},

			options: {
				version: {
					boolean: true,
					description: 'Show the version of particle-cli installed.'
				}
			},

			examples: {
				'$0 setup': 'Set up your Particle account and your device',
				'$0 list': 'Show all your devices and their functions and variables',
				'$0 flash my_device tinker': 'Remotely update your device to run the default Tinker app',
				'$0 call my_device_name digitalwrite D7=HIGH': 'Call a function on your device running Tinker to toggle the onboard LED'
			},

			epilogue: 'For more information, visit our documentation at https://docs.particle.io',

			version: app.showVersion,

			/**
    * Setup global attributes from the parsed arguments.
    * @param {*} yargs The yargs parser to setup
    * @param {CLIRootCommandCategory} root The root command category to setup.
    */
			setup(yargs, root) {
				(0, _cli2.default)({ commandProcessor, root, app });
				app.addGlobalOptions(yargs);
				_lodash2.default.each(root.commands, command => app.addGlobalSetup(command));
			},

			/**
    * Set up the global state from the initial command parsing.
    * @param {*} argv The parsed command line arguments.
    */
			parsed(argv) {
				global.isInteractive = argv.interactive === true || _tty2.default.isatty(_process2.default.stdin) && !argv.nonInteractive;
				global.verboseLevel = argv.verbose + 1 - argv.quiet;
				global.outputJson = argv.json;
			}
		});
	}

	addGlobalSetup(cat) {
		if (!cat.options.setup) {
			cat.options.setup = this.addGlobalOptions.bind(this);
		} else {
			const oldFunc = cat.options.setup;
			cat.options.setup = yargs => {
				oldFunc(yargs);
				this.addGlobalSetup(yargs);
			};
		}
		_lodash2.default.each(cat.commands, command => this.addGlobalSetup(command));
	}

	addGlobalOptions(yargs) {
		// the options are added by each subcommand, so we just
		// todo - if a command overrides an option, then it should not be set as a global option.
		// This is probably best moved into the command-processor.js module
		_lodash2.default.each(this.rootCategory.options.inherited.options, (opt, name) => {
			yargs.group(name, 'Global Options:');
		});
	}

	runCommand(args) {
		const errors = commandProcessor.createErrorHandler();
		this.rootCategory = this.setupCommandProcessor();
		let argv;
		try {
			argv = commandProcessor.parse(this.rootCategory, args);
		} catch (error) {
			error.isUsageError = true;
			errors(error);
		}
		// we want to separate execution from parsing, but yargs wants to execute help/version when parsing args.
		// this also gives us more control.
		// todo - handle root command passing --version
		if (argv.help) {
			commandProcessor.showHelp();
		} else if (argv.clierror) {
			errors(argv.clierror);
		} else if (argv.clicommand) {
			(0, _when2.default)(argv.clicommand.exec(argv)).done(result => result, errors);
		}
	}

	showVersion() {
		let exit = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : true;

		console.log(_package2.default.version);
		_process2.default.exit();
	}

	hasArg(name, args) {
		const index = args.indexOf(name);
		if (index >= 0) {
			args.splice(index, 1);
			return true;
		}
		return false;
	}

	run(args) {
		settings.whichProfile();
		settings.loadOverrides();

		settings.disableUpdateCheck = this.hasArg('--no-update-check', args);
		const force = this.hasArg('--force-update-check', args);

		return (0, _updateCheck2.default)(settings.disableUpdateCheck, force).then(() => {
			const cmdargs = args.slice(2); // remove executable and script
			return this.runCommand(cmdargs);
		}).catch(commandProcessor.createErrorHandler());
	}
}
exports.default = CLI;
module.exports = exports['default'];