'use strict';

Object.defineProperty(exports, "__esModule", {
	value: true
});
exports.command = command;

var _cmd = require('../cmd');

var _particleLibraryManager = require('particle-library-manager');

var _path = require('path');

var _path2 = _interopRequireDefault(_path);

var _log = require('../lib/log');

var _log2 = _interopRequireDefault(_log);

var _chalk = require('chalk');

var _chalk2 = _interopRequireDefault(_chalk);

var _prompts = require('../../dist/lib/prompts');

var _prompts2 = _interopRequireDefault(_prompts);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

// todo - this is pulled from validating_editor in particle-dev-libraries. Please refactor/DRY.

function validationMessage(validationResult, fieldName) {
	return fieldName + ' ' + validationResult.errors[fieldName];
}

function fieldValidator(fieldName) {
	const validator = value => {
		const result = (0, _particleLibraryManager.validateField)(fieldName, value);
		return result && result.valid ? '' : validationMessage(result, fieldName);
	};
	return validator;
}

function yesNoValidator() {
	const validator = value => {
		if (!value || value !== 'Y' && value !== 'y' && value !== 'N' && value !== 'n') {
			return 'Please answer "y" or "n" - you typed ' + value;
		}
	};
	return validator;
}

class CLIProjectInitCommandSite extends _cmd.ProjectInitCommandSite {
	constructor(_ref) {
		let name = _ref.name,
		    directory = _ref.directory;

		super();
		this._name = name;
		this._dir = directory;
	}

	/**
  * Has a dialog with the user. Returns a truthy value if project init should continue.
  * “What would you like to call your project?”
  Then, you are asked to choose where you would like to save your project
  “Would you like to create your project in the default project directory? (Y/n)
  If you answer Y, the project directory is created under /home/particle/projects/myproject, and the CLI replies, “A new project named “myproject” was successfully created in home/particle/projects/myproject
  Then, you are automatically taken to the new project directory in your current session
  If you answer N to the first question, the CLI asks, “Would you like to create your project in <current path>? Type “n” to cancel. (Y/n)
  If you answer Y, the project directory is created under <current path>/myproject
  If you answer anything else, the `project init` process is cancelled
  * @returns {Promise} promise
  */
	dialog() {
		const promptForName = () => {
			return this._name || this.prompt('What would you like to call your project? [myproject]', 'myproject', fieldValidator('name'));
		};

		function isYes(result) {
			return (result || '').toLowerCase() === 'y';
		}

		const promptForLocation = (commonLocation, currentLocation) => {
			return this._dir || this.prompt('Would you like to create your project in the default project directory? [Y/n]', 'Y', yesNoValidator()).then(result => {
				if (isYes(result)) {
					return commonLocation;
				}
			}).then(location => {
				return location || this.prompt(`Would you like to create your project in ${_chalk2.default.bold(currentLocation)}? Type “n” to cancel. [Y/n]`, 'Y', yesNoValidator()).then(result => {
					return isYes(result) ? currentLocation : '';
				});
			});
		};

		return Promise.resolve().then(() => promptForName()).then(name => {
			this._name = name;
			if (name) {
				return promptForLocation(new _cmd.Projects().myProjectsFolder(), process.cwd());
			}
		}).then(directory => {
			this._dir = directory;
			return this._dir && this._name; // are we ready?
		});
	}

	/**
  * Prompts the user to enter a value
  * @param {string} message       The message to prompt with
  * @param {string} value         The default value
     * @param {function(value)}         validator The validator that is used to validate the response.
  * @returns {Promise} promise
  */
	prompt(message, value, validator) {
		return _prompts2.default.promptAndValidate(message, value, validator);
	}

	name() {
		return this._name;
	}

	directory() {
		return _path2.default.join(this._dir, this._name);
	}

	filesystem() {
		return require('fs');
	}

	/**
  * Notification that the directory exists, should creation proceed?
  * @param {String} dir The directory that exists.
  * @returns {boolean} true to continue with project creation
  * The response can be a direct value or a promise. If the promise is falsey then the process is stopped.
  */
	notifyDirectoryExists(dir) {
		// creating a new project is non-destructive so we allow it always
		return true;
	}

	/**
  * Notification of the entire project creation operation.
  * @param {String} path      The directory that will contain the project
  * @param {Promise} promise   The promise to create the project in the given directory
  * @returns {Promise} Promise
  */
	notifyCreatingProject(path, promise) {
		return _log2.default.info(`Initializing project in directory ${_chalk2.default.bold(path)}...`);
	}

	/**
  * Notification that the command is creating a file or directory.
  * @param {String} path          The path being created
  * @param {Promise} promise       The promise to create the path. The implementation may
  * extend this promise and return the new extension. This may be undefined also.
  * @return {Promise} undefined to use the original promise, or a wrapped version of the promise.
  */
	notifyCreatingPath(path, promise) {
		return promise;
	}

	notifyProjectNotCreated(directory) {
		_log2.default.warn('Project initialization was cancelled.');
	}

	notifyProjectCreated(directory) {
		_log2.default.success(`A new project has been initialized in directory ${_chalk2.default.bold(directory)}`);
	}
}

function command(argv) {
	const cmd = new _cmd.ProjectInitCommand();
	const site = new CLIProjectInitCommandSite({
		directory: argv.params.dir,
		name: argv.name
	});
	return site.dialog().then(ready => {
		if (ready) {
			return site.run(cmd);
		}
	});
}