'use strict';

Object.defineProperty(exports, "__esModule", {
	value: true
});
exports.UnauthorizedError = exports.convertApiError = undefined;

var _particleCommands = require('particle-commands');

Object.defineProperty(exports, 'convertApiError', {
	enumerable: true,
	get: function () {
		return _particleCommands.convertApiError;
	}
});

var _fs = require('fs');

var _fs2 = _interopRequireDefault(_fs);

var _when = require('when');

var _when2 = _interopRequireDefault(_when);

var _particleApiJs = require('particle-api-js');

var _particleApiJs2 = _interopRequireDefault(_particleApiJs);

var _log = require('../lib/log');

var _log2 = _interopRequireDefault(_log);

var _lodash = require('lodash');

var _lodash2 = _interopRequireDefault(_lodash);

var _url = require('url');

var _url2 = _interopRequireDefault(_url);

var _chalk = require('chalk');

var _chalk2 = _interopRequireDefault(_chalk);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

class UnauthorizedError extends Error {
	constructor(message) {
		super();
		this.message = message || 'Invalid access token';
		this.name = UnauthorizedError.name;
		if (typeof Error.captureStackTrace === 'function') {
			Error.captureStackTrace(this, UnauthorizedError);
		}
	}
}

class ParticleApi {
	constructor(baseUrl, options) {
		this.api = new _particleApiJs2.default({
			baseUrl: baseUrl,
			clientId: options.clientId || 'particle-cli',
			clientSecret: 'particle-cli',
			tokenDuration: 7776000, // 90 days
			debug: this._debug.bind(this)
		});
		this.accessToken = options.accessToken;
	}

	login(username, password) {
		return (0, _when2.default)(this.api.login({ username: username, password: password })).then(result => {
			this.accessToken = result.body.access_token;
			return this.accessToken;
		});
	}

	logout() {
		this.accessToken = undefined;
		return _when2.default.resolve();
	}

	removeAccessToken(username, password, token) {
		return (0, _when2.default)(this.api.removeAccessToken({ username, password, token })).then(() => {
			if (token === this.accessToken) {
				this.logout();
			}
		});
	}

	listDevices() {
		return this._wrap(this.api.listDevices({ auth: this.accessToken }));
	}

	getDeviceAttributes(deviceId) {
		return this._wrap(this.api.getDevice({ deviceId, auth: this.accessToken }));
	}

	claimDevice(deviceId, requestTransfer) {
		return this._wrap(this.api.claimDevice({ deviceId, requestTransfer, auth: this.accessToken }));
	}

	removeDevice(deviceId) {
		return this._wrap(this.api.removeDevice({ deviceId, auth: this.accessToken }));
	}

	renameDevice(deviceId, name) {
		return this._wrap(this.api.renameDevice({ deviceId, name, auth: this.accessToken }));
	}

	signalDevice(deviceId, signal) {
		return this._wrap(this.api.signalDevice({ deviceId, signal, auth: this.accessToken }));
	}

	listBuildTargets(onlyFeatured) {
		return this._wrap(this.api.listBuildTargets({ onlyFeatured, auth: this.accessToken }));
	}

	compileCode(files, platformId, targetVersion) {
		return this._wrap(this.api.compileCode({ files, platformId, targetVersion, auth: this.accessToken }));
	}

	downloadFirmwareBinary(binaryId, downloadPath) {
		return _when2.default.promise((resolve, reject) => {
			const req = this.api.downloadFirmwareBinary({ binaryId, auth: this.accessToken });
			req.pipe(_fs2.default.createWriteStream(downloadPath)).on('error', reject).on('finish', resolve);
		});
	}

	getEventStream(deviceId, name) {
		return this.api.getEventStream({ deviceId, name, auth: this.accessToken });
	}

	publishEvent(name, data, isPrivate) {
		return this.api.publishEvent({ name, data, isPrivate, auth: this.accessToken });
	}

	_wrap(promise) {
		return (0, _when2.default)(promise).then(result => result.body).catch(this._checkToken);
	}

	_checkToken(err) {
		if (err.statusCode === 401) {
			return _when2.default.reject(new UnauthorizedError());
		}
		return _when2.default.reject(err);
	}

	_debug(req) {
		if (global.verboseLevel > 3) {
			const request = req.url ? req : req.req;
			let destUrl;
			// superagent vs plain http
			if (request.url) {
				const parsedUrl = _url2.default.parse(request.url);
				parsedUrl.query = request.qs;
				destUrl = _url2.default.format(parsedUrl);
			} else {
				destUrl = `${request.path}`;
			}
			_log2.default.silly(_chalk2.default.underline('REQUEST'));
			_log2.default.silly(`${request.method.toUpperCase()} ${destUrl}`);
			const headers = request.header || request._headers;
			if (headers && Object.keys(headers).length) {
				_log2.default.silly(_lodash2.default.map(headers, (v, k) => {
					let val = v;
					if (k === 'authorization') {
						val = val.replace(this.accessToken, '<redacted>');
					}
					return `${k}: ${val}`;
				}).join('\n'));
			}
			if (request._data && Object.keys(request._data).length) {
				const clonedData = Object.assign({}, request._data);
				if (clonedData.password) {
					clonedData.password = '<redacted>';
				}
				_log2.default.silly(clonedData);
			}
			_log2.default.silly();
			req.on('response', res => {
				_log2.default.silly(_chalk2.default.underline('RESPONSE'));
				_log2.default.silly(`${request.method.toUpperCase()} ${destUrl}`);
				_log2.default.silly(res.statusCode);
				if (res.text || res.body) {
					_log2.default.silly(res.text || res.body);
				}
				_log2.default.silly();
			});
		}
	}
}

exports.default = ParticleApi;
exports.UnauthorizedError = UnauthorizedError;