'use strict';

const VError = require('verror');
const when = require('when');

const prompt = require('inquirer').prompt;
const fs = require('fs');

const ApiClient = require('../lib/ApiClient');
const utilities = require('../lib/utilities');

class WebhookCommand {
	constructor(options) {
		this.options = options;
	}

	createPOSTHook(_ref) {
		let eventName = _ref.eventName,
		    url = _ref.url,
		    device = _ref.device;

		return this._createHook({ eventName, url, deviceID: device, requestType: 'POST' });
	}

	createGETHook(_ref2) {
		let eventName = _ref2.eventName,
		    url = _ref2.url,
		    device = _ref2.device;

		return this._createHook({ eventName, url, deviceID: device, requestType: 'GET' });
	}

	createHook(_ref3) {
		let eventName = _ref3.eventName,
		    url = _ref3.url,
		    device = _ref3.device,
		    requestType = _ref3.requestType;

		return this._createHook({ eventName, url, deviceID: device, requestType });
	}

	_createHook(_ref4) {
		let eventName = _ref4.eventName,
		    url = _ref4.url,
		    deviceID = _ref4.deviceID,
		    requestType = _ref4.requestType;

		const api = new ApiClient();
		api.ensureToken();

		//if they gave us one thing, and it happens to be a file, and we could parse it as json
		let data = {};
		//particle webhook create xxx.json
		if (eventName && !url && !deviceID) {
			const filename = eventName;

			if (utilities.getFilenameExt(filename).toLowerCase() === '.json') {
				if (!fs.existsSync(filename)) {
					throw new VError(filename + ' is not found.');
				}

				data = utilities.tryParse(fs.readFileSync(filename));
				if (!data) {
					throw new VError('Please check your .json file for syntax error.');
				}
				console.log('Using settings from the file ' + filename);
				//only override these when we didn't get them from the command line
				eventName = data.event || data.eventName;
				url = data.url;
				deviceID = data.deviceid;
			}
		}

		//required param
		if (!eventName) {
			throw new VError('Please specify an event name');
		}

		//required param
		if (!url) {
			throw new VError('Please specify a url');
		}

		const webhookData = Object.assign({
			event: eventName,
			url: url,
			deviceid: deviceID,
			requestType: requestType || data.requestType
		}, data);

		return api.createWebhookWithObj(webhookData).catch(err => {
			throw new VError(api.normalizedApiError(err), 'Error creating webhook');
		});
	}

	deleteHook(_ref5) {
		let hookId = _ref5.hookId;

		const api = new ApiClient();
		api.ensureToken();

		return Promise.resolve().then(() => {
			if (hookId === 'all') {
				// delete all hook using `particle webhook delete all`
				return prompt([{
					type: 'confirm',
					name: 'deleteAll',
					message: 'Do you want to delete ALL your webhooks?',
					default: false
				}]).then(answer => {
					if (answer.deleteAll) {
						return api.listWebhooks().then(hooks => {
							console.log('Found ' + hooks.length + ' hooks registered\n');
							return when.reduce(hooks, (ignore, hook) => {
								console.log('deleting ' + hook.id);
								return api.deleteWebhook(hook.id);
							});
						});
					}
				});
			} else {
				// delete a hook based on ID
				return api.deleteWebhook(hookId);
			}
		}).catch(err => {
			throw new VError(api.normalizedApiError(err), 'Error deleting webhook');
		});
	}

	listHooks() {
		const api = new ApiClient();
		api.ensureToken();

		return api.listWebhooks().then(hooks => {
			console.log('Found ' + hooks.length + ' hooks registered\n');
			for (let i = 0; i < hooks.length; i++) {
				const hook = hooks[i];
				const line = ['    ', i + 1, '.) Hook ID ' + hook.id + ' is watching for ', '"' + hook.event + '"', '\n       ', ' and sending to: ' + hook.url, hook.deviceID ? '\n       ' + ' for device ' + hook.deviceID : '', '\n       ', ' created at ' + hook.created_at, '\n'].join('');

				console.log(line);
			}
		}).catch(err => {
			throw new VError(api.normalizedApiError(err), 'Error listing webhooks');
		});
	}
}

module.exports = WebhookCommand;